/* java.lang.reflect.Field - reflection of Java fields
   Copyright (C) 1998, 2001 Free Software Foundation, Inc.

This file is part of GNU Classpath.

GNU Classpath is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.
 
GNU Classpath is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Classpath; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version. */


package java.lang.reflect;
import gnu.classpath.VMStackWalker;

/**
 * The Field class represents a member variable of a class. It also allows
 * dynamic access to a member, via reflection. This works for both
 * static and instance fields. Operations on Field objects know how to
 * do widening conversions, but throw {@link IllegalArgumentException} if
 * a narrowing conversion would be necessary. You can query for information
 * on this Field regardless of location, but get and set access may be limited
 * by Java language access controls. If you can't do it in the compiler, you
 * can't normally do it here either.<p>
 *
 * <B>Note:</B> This class returns and accepts types as Classes, even
 * primitive types; there are Class types defined that represent each
 * different primitive type.  They are <code>java.lang.Boolean.TYPE,
 * java.lang.Byte.TYPE,</code>, also available as <code>boolean.class,
 * byte.class</code>, etc.  These are not to be confused with the
 * classes <code>java.lang.Boolean, java.lang.Byte</code>, etc., which are
 * real classes.<p>
 *
 * Also note that this is not a serializable class.  It is entirely feasible
 * to make it serializable using the Externalizable interface, but this is
 * on Sun, not me.
 *
 * @author John Keiser
 * @author Eric Blake <ebb9@email.byu.edu>
 * @see Member
 * @see Class
 * @see Class#getField(String)
 * @see Class#getDeclaredField(String)
 * @see Class#getFields()
 * @see Class#getDeclaredFields()
 * @since 1.1
 * @status updated to 1.4
 */

/**
 * Modified by HY and JT for OVM (11/21/2003)
 * and DH
 **/
import org.ovmj.java.Opaque;

public final class Field extends AccessibleObject implements Member {

    final Opaque vmField; // reference to our ED Field object
    private final Class declaringClass;
    private final int modifiers;
    private volatile String name;  // lazily initialized
    private volatile Class type;   // lazily initialized
 
    private volatile boolean declaringClassInitialized = false;
    
    /**
     * This class is uninstantiable except by the OVM
     */
    private Field(Opaque vmField, Class declaringClass, int modifiers) {

        if (vmField == null || declaringClass == null) 
            throw new InternalError("Field constructor got null arg");

	this.vmField = vmField;
	this.declaringClass = declaringClass;
        this.modifiers = modifiers;
    }
    
    /**
     * Gets the class that declared this field, or the class where this field
     * is a non-inherited member.
     * @return the class that declared this member
     */
    public Class getDeclaringClass() {
        return declaringClass;
    }

    /**
     * Gets the name of this field.
     * @return the name of this field
     */
    public String getName() {
        if (name == null)
            name = LibraryImports.nameForField(vmField);
        return name;
    }

  /**
   * Gets the modifiers this field uses.  Use the <code>Modifier</code>
   * class to interpret the values.  A field can only have a subset of the
   * following modifiers: public, private, protected, static, final,
   * transient, and volatile.
   *
   * @return an integer representing the modifiers to this Member
   * @see Modifier
   */
    public int getModifiers() {
	return modifiers;
    }

  /**
   * Gets the type of this field.
   * @return the type of this field
   */
    public Class getType() {
        if (type == null)
            type = LibraryImports.typeForField(vmField);
        return type;
    }

  /**
   * Compare two objects to see if they are semantically equivalent.
   * Two Fields are semantically equivalent if they have the same declaring
   * class, name, and type. Since you can't creat a Field except through
   * the VM, this is just the == relation.
   *
   * @param o the object to compare to
   * @return <code>true</code> if they are equal; <code>false</code> if not
   */
  public boolean equals(Object o)
  {
    return this == o;
  }

  /**
   * Get the hash code for the Field. The Field hash code is the hash code
   * of its name XOR'd with the hash code of its class name.
   *
   * @return the hash code for the object.
   */
  public int hashCode()
  {
    return getDeclaringClass().getName().hashCode() ^ getName().hashCode();
  }

  /**
   * Get a String representation of the Field. A Field's String
   * representation is "&lt;modifiers&gt; &lt;type&gt;
   * &lt;class&gt;.&lt;fieldname&gt;".<br> Example:
   * <code>public transient boolean gnu.parse.Parser.parseComplete</code>
   *
   * @return the String representation of the Field
   */
  public String toString()
  {
    // 64 is a reasonable buffer initial size for field
    StringBuffer sb = new StringBuffer(64);
    Modifier.toString(getModifiers(), sb).append(' ');
    sb.append(getType().getName()).append(' ');
    sb.append(getDeclaringClass().getName()).append('.');
    sb.append(getName());
    return sb.toString();
  }
 
  /**
   * Get the value of this Field.  If it is primitive, it will be wrapped
   * in the appropriate wrapper type (boolean = java.lang.Boolean).<p>
   *
   * If the field is static, <code>o</code> will be ignored. Otherwise, if
   * <code>o</code> is null, you get a <code>NullPointerException</code>,
   * and if it is incompatible with the declaring class of the field, you
   * get an <code>IllegalArgumentException</code>.<p>
   *
   * Next, if this Field enforces access control, your runtime context is
   * evaluated, and you may have an <code>IllegalAccessException</code> if
   * you could not access this field in similar compiled code. If the field
   * is static, and its class is uninitialized, you trigger class
   * initialization, which may end in a
   * <code>ExceptionInInitializerError</code>.<p>
   *
   * Finally, the field is accessed, and primitives are wrapped (but not
   * necessarily in new objects). This method accesses the field of the
   * declaring class, even if the instance passed in belongs to a subclass
   * which declares another field to hide this one.
   *
   * @param o the object to get the value of this Field from
   * @return the value of the Field
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if <code>o</code> is not an instance of
   *         the class or interface declaring this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #getBoolean(Object)
   * @see #getByte(Object)
   * @see #getChar(Object)
   * @see #getShort(Object)
   * @see #getInt(Object)
   * @see #getLong(Object)
   * @see #getFloat(Object)
   * @see #getDouble(Object)
   */
    public Object get(Object o) throws IllegalAccessException{
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }

	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        Class clazz = getType();
	if (clazz == Integer.TYPE)
	    return new Integer(LibraryImports.fieldGetInt(vmField, o));
	else if (clazz == Boolean.TYPE)
	    return new Boolean(LibraryImports.fieldGetBoolean(vmField, o));
	else if (clazz == Byte.TYPE)
	    return new Byte(LibraryImports.fieldGetByte(vmField, o));
	else if (clazz == Character.TYPE)
	    return new Character(LibraryImports.fieldGetChar(vmField, o));
	else if (clazz == Short.TYPE)
	    return new Short(LibraryImports.fieldGetShort(vmField, o));
	else if (clazz == Float.TYPE)
	    return new Float(LibraryImports.fieldGetFloat(vmField, o));
	else if (clazz == Double.TYPE)
	    return new Double(LibraryImports.fieldGetDouble(vmField, o));
	else if (clazz == Long.TYPE)
	    return new Long(LibraryImports.fieldGetLong(vmField, o));
	else {
            AccessChecks.checkReceiver(this, o);
	    return LibraryImports.fieldGetReference(vmField, o);
        }
    }

  /**
   * Get the value of this boolean Field. If the field is static,
   * <code>o</code> will be ignored.
   *
   * @param o the object to get the value of this Field from
   * @return the value of the Field
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a boolean field of
   *         <code>o</code>, or if <code>o</code> is not an instance of the
   *         declaring class of this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #get(Object)
   */
    public boolean getBoolean(Object o) throws IllegalAccessException{
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz != Boolean.TYPE)
            throw new IllegalArgumentException("not a boolean field");
        return LibraryImports.fieldGetBoolean(vmField, o);
    }

  /**
   * Get the value of this byte Field. If the field is static,
   * <code>o</code> will be ignored.
   *
   * @param o the object to get the value of this Field from
   * @return the value of the Field
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a byte field of
   *         <code>o</code>, or if <code>o</code> is not an instance of the
   *         declaring class of this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #get(Object)
   */
    public byte getByte(Object o) throws IllegalAccessException{
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz != Byte.TYPE)
            throw new IllegalArgumentException("not a byte field");
	return LibraryImports.fieldGetByte(vmField, o);
    }

  /**
   * Get the value of this Field as a char. If the field is static,
   * <code>o</code> will be ignored.
   *
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a char field of
   *         <code>o</code>, or if <code>o</code> is not an instance
   *         of the declaring class of this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #get(Object)
   */
    public char getChar(Object o) throws IllegalAccessException{
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz != Character.TYPE)
            throw new IllegalArgumentException("not a char field");
	return LibraryImports.fieldGetChar(vmField, o);
    }

  /**
   * Get the value of this Field as a short. If the field is static,
   * <code>o</code> will be ignored.
   *
   * @param o the object to get the value of this Field from
   * @return the value of the Field
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a byte or short
   *         field of <code>o</code>, or if <code>o</code> is not an instance
   *         of the declaring class of this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #get(Object)
   */
    public short getShort(Object o) throws IllegalAccessException{
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Short.TYPE)
            return LibraryImports.fieldGetShort(vmField, o);
        else if (clazz == Byte.TYPE)
            return LibraryImports.fieldGetByte(vmField, o);
        else
            throw new IllegalArgumentException("not a short compatible field");
    }

  /**
   * Get the value of this Field as an int. If the field is static,
   * <code>o</code> will be ignored.
   *
   * @param o the object to get the value of this Field from
   * @return the value of the Field
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a byte, short, char, or
   *         int field of <code>o</code>, or if <code>o</code> is not an
   *         instance of the declaring class of this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #get(Object)
   */
    public int getInt(Object o) throws IllegalAccessException{
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Integer.TYPE)
            return LibraryImports.fieldGetInt(vmField, o);
        else if (clazz == Character.TYPE)
            return LibraryImports.fieldGetChar(vmField, o);
        else if (clazz == Short.TYPE)
            return LibraryImports.fieldGetShort(vmField, o);
        else if (clazz == Byte.TYPE)
            return LibraryImports.fieldGetByte(vmField, o);
        else
            throw new IllegalArgumentException("not an int compatible field");
    }

  /**
   * Get the value of this Field as a long. If the field is static,
   * <code>o</code> will be ignored.
   *
   * @param o the object to get the value of this Field from
   * @return the value of the Field
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a byte, short, char, int,
   *         or long field of <code>o</code>, or if <code>o</code> is not an
   *         instance of the declaring class of this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #get(Object)
   */
    public long getLong(Object o) throws IllegalAccessException{
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Long.TYPE)
            return LibraryImports.fieldGetLong(vmField, o);
        else if (clazz == Integer.TYPE)
            return LibraryImports.fieldGetInt(vmField, o);
        else if (clazz == Character.TYPE)
            return LibraryImports.fieldGetChar(vmField, o);
        else if (clazz == Short.TYPE)
            return LibraryImports.fieldGetShort(vmField, o);
        else if (clazz == Byte.TYPE)
            return LibraryImports.fieldGetByte(vmField, o);
        else
            throw new IllegalArgumentException("not a long compatible field");
    }

  /**
   * Get the value of this Field as a float. If the field is static,
   * <code>o</code> will be ignored.
   *
   * @param o the object to get the value of this Field from
   * @return the value of the Field
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a byte, short, char, int,
   *         long, or float field of <code>o</code>, or if <code>o</code> is
   *         not an instance of the declaring class of this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #get(Object)
   */
    public float getFloat(Object o) throws IllegalAccessException{
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Float.TYPE)
	return LibraryImports.fieldGetFloat(vmField, o);
        else if (clazz == Long.TYPE)
            return LibraryImports.fieldGetLong(vmField, o);
        else if (clazz == Integer.TYPE)
            return LibraryImports.fieldGetInt(vmField, o);
        else if (clazz == Character.TYPE)
            return LibraryImports.fieldGetChar(vmField, o);
        else if (clazz == Short.TYPE)
            return LibraryImports.fieldGetShort(vmField, o);
        else if (clazz == Byte.TYPE)
            return LibraryImports.fieldGetByte(vmField, o);
        else
            throw new IllegalArgumentException("not a float compatible field");
    }

  /**
   * Get the value of this Field as a double. If the field is static,
   * <code>o</code> will be ignored.
   *
   * @param o the object to get the value of this Field from
   * @return the value of the Field
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a byte, short, char, int,
   *         long, float, or double field of <code>o</code>, or if
   *         <code>o</code> is not an instance of the declaring class of this
   *         field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #get(Object)
   */
    public double getDouble(Object o) throws IllegalAccessException{
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Double.TYPE)
            return LibraryImports.fieldGetDouble(vmField, o);
        else if (clazz == Float.TYPE)
            return LibraryImports.fieldGetFloat(vmField, o);
        else if (clazz == Long.TYPE)
            return LibraryImports.fieldGetLong(vmField, o);
        else if (clazz == Integer.TYPE)
            return LibraryImports.fieldGetInt(vmField, o);
        else if (clazz == Character.TYPE)
            return LibraryImports.fieldGetChar(vmField, o);
        else if (clazz == Short.TYPE)
            return LibraryImports.fieldGetShort(vmField, o);
        else if (clazz == Byte.TYPE)
            return LibraryImports.fieldGetByte(vmField, o);
        else
            throw new IllegalArgumentException("not a double compatible field");    }

  /**
   * Set the value of this Field.  If it is a primitive field, the value
   * will be unwrapped from the passed object (boolean = java.lang.Boolean).<p>
   *
   * If the field is static, <code>o</code> will be ignored. Otherwise, if
   * <code>o</code> is null, you get a <code>NullPointerException</code>,
   * and if it is incompatible with the declaring class of the field, you
   * get an <code>IllegalArgumentException</code>.<p>
   *
   * Next, if this Field enforces access control, your runtime context is
   * evaluated, and you may have an <code>IllegalAccessException</code> if
   * you could not access this field in similar compiled code. This also
   * occurs whether or not there is access control if the field is final.
   * If the field is primitive, and unwrapping your argument fails, you will
   * get an <code>IllegalArgumentException</code>; likewise, this error
   * happens if <code>value</code> cannot be cast to the correct object type.
   * If the field is static, and its class is uninitialized, you trigger class
   * initialization, which may end in a
   * <code>ExceptionInInitializerError</code>.<p>
   *
   * Finally, the field is set with the widened value. This method accesses
   * the field of the declaring class, even if the instance passed in belongs
   * to a subclass which declares another field to hide this one.
   *
   * @param o the object to set this Field on
   * @param value the value to set this Field to
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if <code>value</code> cannot be
   *         converted by a widening conversion to the underlying type of
   *         the Field, or if <code>o</code> is not an instance of the class
   *         declaring this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #setBoolean(Object, boolean)
   * @see #setByte(Object, byte)
   * @see #setChar(Object, char)
   * @see #setShort(Object, short)
   * @see #setInt(Object, int)
   * @see #setLong(Object, long)
   * @see #setFloat(Object, float)
   * @see #setDouble(Object, double)
   */
    public void set(Object o, Object value) throws IllegalAccessException {
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
	checkAssignment();
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
	AccessChecks.checkReceiver(this, o);
        Class clazz = getType();
        if (clazz.isPrimitive()) {
            try {
                // boolean can only be set by Boolean
                if (clazz == Boolean.TYPE) {
                    LibraryImports.fieldSetBoolean(vmField, o, ((Boolean)value).booleanValue());
                    return;
                }
                // character can be set only by Character
                else if (clazz == Character.TYPE) {
                    LibraryImports.fieldSetChar(vmField, o, ((Character)value).charValue());
                    return;
                }
                // for Number types, widening conversions are allowed
                else if (clazz == Byte.TYPE) {
                    LibraryImports.fieldSetByte(vmField, o, ((Byte)value).byteValue());
                    return;
                }
                else if (clazz == Short.TYPE) {
                    if (value instanceof Byte || value instanceof Short) {
                        LibraryImports.fieldSetShort(vmField, o, ((Number)value).shortValue());
                        return;
                    }
                }
                else if (clazz == Integer.TYPE) {
                    if (value instanceof Byte || value instanceof Short ||
                        value instanceof Integer) {
                        LibraryImports.fieldSetInt(vmField, o, ((Number)value).intValue());
                        return;
                    }
                    else if (value instanceof Character) {
                        LibraryImports.fieldSetInt(vmField, o, ((Character)value).charValue());
                        return;
                    }
                }
                else if (clazz == Long.TYPE) {
                    if (value instanceof Byte || value instanceof Short ||
                        value instanceof Integer || value instanceof Long) {
                        LibraryImports.fieldSetLong(vmField, o, ((Number)value).longValue());
                        return;
                    }
                    else if (value instanceof Character) {
                        LibraryImports.fieldSetLong(vmField, o, ((Character)value).charValue());
                        return;
                    }
                }
                else if (clazz == Float.TYPE) {
                    if (value instanceof Byte || value instanceof Short ||
                        value instanceof Integer || value instanceof Long ||
                        value instanceof Float) {
                        LibraryImports.fieldSetFloat(vmField, o, ((Number)value).floatValue());
                        return;
                    }
                    else if (value instanceof Character) {
                        LibraryImports.fieldSetFloat(vmField, o, ((Character)value).charValue());
                        return;
                    }
                }
                else if (clazz == Double.TYPE) {
                    if (value instanceof Byte || value instanceof Short ||
                        value instanceof Integer || value instanceof Long ||
                        value instanceof Float || value instanceof Double) {
                        LibraryImports.fieldSetDouble(vmField, o, ((Number)value).doubleValue());
                        return;
                    }
                    else if (value instanceof Character) {
                        LibraryImports.fieldSetDouble(vmField, o, ((Character)value).charValue());
                        return;
                    }

                }

                // wrong type passed or null
                throw new IllegalArgumentException("wrong argument type (or null) for parameter");
            }
            catch (ClassCastException cce) {
                throw new IllegalArgumentException("wrong argument type for parameter");
            }
        }
        else {
            if (value != null) checkValueType(value.getClass());
	    LibraryImports.fieldSetReference(vmField, o, value);
        }
    }

  /**
   * Set this boolean Field. If the field is static, <code>o</code> will be
   * ignored.
   *
   * @param o the object to set this Field on
   * @param value the value to set this Field to
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a boolean field, or if
   *         <code>o</code> is not an instance of the class declaring this
   *         field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #set(Object, Object)
   */
    public void setBoolean(Object o, boolean value) throws IllegalAccessException {
	checkAssignment();
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Boolean.TYPE)
            LibraryImports.fieldSetBoolean(vmField, o, value);
        else
            throw new IllegalArgumentException("not a boolean field");
    }

  /**
   * Set this byte Field. If the field is static, <code>o</code> will be
   * ignored.
   *
   * @param o the object to set this Field on
   * @param value the value to set this Field to
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a byte, short, int, long,
   *         float, or double field, or if <code>o</code> is not an instance
   *         of the class declaring this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #set(Object, Object)
   */
  public void setByte(Object o, byte value) throws IllegalAccessException {
	checkAssignment();
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Byte.TYPE)
            LibraryImports.fieldSetByte(vmField, o, value);
        else if (clazz == Short.TYPE)
            LibraryImports.fieldSetShort(vmField, o, value);
        else if (clazz == Integer.TYPE)
            LibraryImports.fieldSetInt(vmField, o, value);
        else if (clazz == Long.TYPE)
            LibraryImports.fieldSetLong(vmField, o, value);
        else if (clazz == Float.TYPE)
            LibraryImports.fieldSetFloat(vmField, o, value);
        else if (clazz == Double.TYPE)
            LibraryImports.fieldSetDouble(vmField, o, value);
        else
            throw new IllegalArgumentException("not a byte compatible field");
  }
  /**
   * Set this char Field. If the field is static, <code>o</code> will be
   * ignored.
   *
   * @param o the object to set this Field on
   * @param value the value to set this Field to
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a char, int, long,
   *         float, or double field, or if <code>o</code> is not an instance
   *         of the class declaring this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #set(Object, Object)
   */
    public void setChar(Object o, char value) throws IllegalAccessException {
	checkAssignment();
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Character.TYPE)
            LibraryImports.fieldSetChar(vmField, o, value);
        else if (clazz == Integer.TYPE)
            LibraryImports.fieldSetInt(vmField, o, value);
        else if (clazz == Long.TYPE)
            LibraryImports.fieldSetLong(vmField, o, value);
        else if (clazz == Float.TYPE)
            LibraryImports.fieldSetFloat(vmField, o, value);
        else if (clazz == Double.TYPE)
            LibraryImports.fieldSetDouble(vmField, o, value);
        else
            throw new IllegalArgumentException("not a char compatible field");

    }

  /**
   * Set this short Field. If the field is static, <code>o</code> will be
   * ignored.
   *
   * @param o the object to set this Field on
   * @param value the value to set this Field to
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a short, int, long,
   *         float, or double field, or if <code>o</code> is not an instance
   *         of the class declaring this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #set(Object, Object)
   */
    public void setShort(Object o, short value) throws IllegalAccessException {
	checkAssignment();
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Short.TYPE)
            LibraryImports.fieldSetShort(vmField, o, value);
        else if (clazz == Integer.TYPE)
            LibraryImports.fieldSetInt(vmField, o, value);
        else if (clazz == Long.TYPE)
            LibraryImports.fieldSetLong(vmField, o, value);
        else if (clazz == Float.TYPE)
            LibraryImports.fieldSetFloat(vmField, o, value);
        else if (clazz == Double.TYPE)
            LibraryImports.fieldSetDouble(vmField, o, value);
        else
            throw new IllegalArgumentException("not a short compatible field");

    }

  /**
   * Set this int Field. If the field is static, <code>o</code> will be
   * ignored.
   *
   * @param o the object to set this Field on
   * @param value the value to set this Field to
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not an int, long, float, or
   *         double field, or if <code>o</code> is not an instance of the
   *         class declaring this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #set(Object, Object)
   */
    public void setInt(Object o, int value) throws IllegalAccessException {
	checkAssignment();
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Integer.TYPE)
            LibraryImports.fieldSetInt(vmField, o, value);
        else if (clazz == Long.TYPE)
            LibraryImports.fieldSetLong(vmField, o, value);
        else if (clazz == Float.TYPE)
            LibraryImports.fieldSetFloat(vmField, o, value);
        else if (clazz == Double.TYPE)
            LibraryImports.fieldSetDouble(vmField, o, value);
        else
            throw new IllegalArgumentException("not an int compatible field");
    }

  /**
   * Set this long Field. If the field is static, <code>o</code> will be
   * ignored.
   *
   * @param o the object to set this Field on
   * @param value the value to set this Field to
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a long, float, or double
   *         field, or if <code>o</code> is not an instance of the class
   *         declaring this field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #set(Object, Object)
   */
    public void setLong(Object o, long value) throws IllegalAccessException {
	checkAssignment();
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Long.TYPE)
            LibraryImports.fieldSetLong(vmField, o, value);
        else if (clazz == Float.TYPE)
            LibraryImports.fieldSetFloat(vmField, o, value);
        else if (clazz == Double.TYPE)
            LibraryImports.fieldSetDouble(vmField, o, value);
        else
            throw new IllegalArgumentException("not a long compatible field");
    }

  /**
   * Set this float Field. If the field is static, <code>o</code> will be
   * ignored.
   *
   * @param o the object to set this Field on
   * @param value the value to set this Field to
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a float or long field, or
   *         if <code>o</code> is not an instance of the class declaring this
   *         field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #set(Object, Object)
   */
    public void setFloat(Object o, float value) throws IllegalAccessException {
	checkAssignment();
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Float.TYPE)
            LibraryImports.fieldSetFloat(vmField, o, value);
        else if (clazz == Double.TYPE)
            LibraryImports.fieldSetDouble(vmField, o, value);
        else
            throw new IllegalArgumentException("not a float compatible field");
    }

  /**
   * Set this double Field. If the field is static, <code>o</code> will be
   * ignored.
   *
   * @param o the object to set this Field on
   * @param value the value to set this Field to
   * @throws IllegalAccessException if you could not normally access this field
   *         (i.e. it is not public)
   * @throws IllegalArgumentException if this is not a double field, or if
   *         <code>o</code> is not an instance of the class declaring this
   *         field
   * @throws NullPointerException if <code>o</code> is null and this field
   *         requires an instance
   * @throws ExceptionInInitializerError if accessing a static field triggered
   *         class initialization, which then failed
   * @see #set(Object, Object)
   */
    public void setDouble(Object o, double value) throws IllegalAccessException {
	checkAssignment();
	AccessChecks.checkAccessible(VMStackWalker.getCallingClass(),
				     this, o);
        AccessChecks.checkReceiver(this, o);
        // our declaring class must be initialized before a field
        // can be got or set
        if (!declaringClassInitialized) {
            LibraryImports.initializeClass(declaringClass);
            declaringClassInitialized = true;
        }
        Class clazz = getType();
        if (clazz == Double.TYPE)
            LibraryImports.fieldSetDouble(vmField, o, value);
        else
            throw new IllegalArgumentException("not a double compatible field");
    }


    void checkValueType(Class valClass) {
        Class clazz = getType();
        if (clazz != valClass && !clazz.isAssignableFrom(valClass))
            throw new IllegalArgumentException("wrong argument type");
    }

    void checkAssignment() throws IllegalAccessException {
	if (Modifier.isFinal(modifiers) && !isAccessible())
	    throw new IllegalAccessException("attempt to write final field");
    }
}
